{% extends "_base.md.j2" %}
{% block content %}
## 🟡 Schema Drift Detected

{{ message }}

| Detail | Value |
|:-------|:------|
| Run ID | `{{ run_id }}` |
| Engine | {{ engine }} |
| Source | `{{ source_path or 'N/A' }}` |
| Time (UTC) | {{ timestamp_utc }} |
{% if contract.domain %}| Domain | {{ contract.domain }} |
{% endif %}{% if contract.system %}| System | {{ contract.system }} |
{% endif %}

> Review the contract schema and update field definitions if the drift is intentional.
{% endblock %}
